/*
 * Copyright (C) 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.car.notification;

import static com.google.common.truth.Truth.assertThat;

import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

import android.app.Notification;
import android.content.Context;
import android.content.Intent;
import android.os.UserHandle;
import android.service.notification.StatusBarNotification;
import android.telephony.TelephonyManager;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InOrder;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.robolectric.RobolectricTestRunner;
import org.robolectric.RuntimeEnvironment;

import java.util.ArrayList;
import java.util.List;

@RunWith(RobolectricTestRunner.class)
public class PreprocessingManagerTest {

    private Context mContext;

    private static final String PKG = "com.package.blah";
    private static final String OP_PKG = "OpPackage";
    private static final int ID = 1;
    private static final String TAG = "Tag";
    private static final int UID = 2;
    private static final int INITIAL_PID = 3;
    private static final String CHANNEL_ID = "CHANNEL_ID";
    private static final String CONTENT_TITLE = "CONTENT_TITLE";
    private static final String OVERRIDE_GROUP_KEY = "OVERRIDE_GROUP_KEY";
    private static final long POST_TIME = 12345l;
    private static final UserHandle USER_HANDLE = new UserHandle(12);

    private PreprocessingManager mPreprocessingManager;

    @Mock
    private PreprocessingManager.CallStateListener mCallStateListener1;
    @Mock
    private PreprocessingManager.CallStateListener mCallStateListener2;

    @Before
    public void setupBaseActivityAndLayout() {
        MockitoAnnotations.initMocks(this);
        mContext = RuntimeEnvironment.application;
        mPreprocessingManager = PreprocessingManager.getInstance(mContext);

        // Always start system with no phone calls in progress.
        Intent intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_IDLE);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);
    }

    @Test
    public void group_autoGeneratedGroupWithNoGroupChildren_groupSummaryShouldNotShow() {
        List<StatusBarNotification> list = new ArrayList<>();
        list.add(getEmptyAutoGeneratedGroupSummary());
        List<NotificationGroup> groupResult = mPreprocessingManager.group(list);
        assertThat(groupResult.size() == 0).isTrue();
    }

    @Test
    public void addCallStateListener_preCall_triggerChanges() {
        InOrder listenerInOrder = Mockito.inOrder(mCallStateListener1);
        mPreprocessingManager.addCallStateListener(mCallStateListener1);
        listenerInOrder.verify(mCallStateListener1).onCallStateChanged(false);

        Intent intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_OFFHOOK);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        listenerInOrder.verify(mCallStateListener1).onCallStateChanged(true);
    }

    @Test
    public void addCallStateListener_midCall_triggerChanges() {
        Intent intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_OFFHOOK);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        mPreprocessingManager.addCallStateListener(mCallStateListener1);

        verify(mCallStateListener1).onCallStateChanged(true);
    }

    @Test
    public void addCallStateListener_postCall_triggerChanges() {
        Intent intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_OFFHOOK);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_IDLE);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        mPreprocessingManager.addCallStateListener(mCallStateListener1);

        verify(mCallStateListener1).onCallStateChanged(false);
    }

    @Test
    public void addSameCallListenerTwice_dedupedCorrectly() {
        mPreprocessingManager.addCallStateListener(mCallStateListener1);

        verify(mCallStateListener1).onCallStateChanged(false);
        mPreprocessingManager.addCallStateListener(mCallStateListener1);

        verify(mCallStateListener1, times(1)).onCallStateChanged(false);
    }

    @Test
    public void removeCallStateListener_midCall_triggerChanges() {
        Intent intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_OFFHOOK);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        mPreprocessingManager.addCallStateListener(mCallStateListener1);
        // Should get triggered with true before calling removeCallStateListener
        mPreprocessingManager.removeCallStateListener(mCallStateListener1);

        intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_IDLE);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        verify(mCallStateListener1, never()).onCallStateChanged(false);
    }

    @Test
    public void multipleCallStateListeners_triggeredAppropriately() {
        InOrder listenerInOrder1 = Mockito.inOrder(mCallStateListener1);
        InOrder listenerInOrder2 = Mockito.inOrder(mCallStateListener2);
        mPreprocessingManager.addCallStateListener(mCallStateListener1);
        listenerInOrder1.verify(mCallStateListener1).onCallStateChanged(false);

        Intent intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_OFFHOOK);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        mPreprocessingManager.addCallStateListener(mCallStateListener2);
        mPreprocessingManager.removeCallStateListener(mCallStateListener1);

        listenerInOrder1.verify(mCallStateListener1).onCallStateChanged(true);
        listenerInOrder2.verify(mCallStateListener2).onCallStateChanged(true);

        intent = new Intent(TelephonyManager.ACTION_PHONE_STATE_CHANGED);
        intent.putExtra(TelephonyManager.EXTRA_STATE, TelephonyManager.EXTRA_STATE_IDLE);
        mPreprocessingManager.mIntentReceiver.onReceive(mContext, intent);

        // only listener 2 should be triggered w/ false
        listenerInOrder1.verifyNoMoreInteractions();
        listenerInOrder2.verify(mCallStateListener2).onCallStateChanged(false);
    }

    private StatusBarNotification getEmptyAutoGeneratedGroupSummary() {
        Notification notification = new Notification.Builder(mContext, CHANNEL_ID)
                .setContentTitle(CONTENT_TITLE)
                .setSmallIcon(android.R.drawable.sym_def_app_icon)
                .setGroup(OVERRIDE_GROUP_KEY)
                .setGroupSummary(true)
                .build();
        StatusBarNotification statusBarNotification = new StatusBarNotification(
                PKG, OP_PKG, ID, TAG, UID, INITIAL_PID, notification, USER_HANDLE,
                OVERRIDE_GROUP_KEY, POST_TIME);
        statusBarNotification.setOverrideGroupKey(OVERRIDE_GROUP_KEY);

        return statusBarNotification;
    }
}
